import { getTangent } from '../geomutils';
import ConstraintAdder from './adder';
import { incidence, horizontality, tangency, verticality, fixation } from './type';
import * as THREE from 'three';

var ConstraintGenerator =
/*#__PURE__*/
function () {
  function ConstraintGenerator(sketcher) {
    this.sketch = sketcher.activeSketch;
    this.adder = new ConstraintAdder(sketcher);
    this.generate = {
      horizontality: true,
      verticality: true,
      pointCoincidence: true,
      tangency: true
    };
  } // generate implied constraints which are somehow related to "relatedTo" object


  var _proto = ConstraintGenerator.prototype;

  _proto.generateImpliedConstraints = function generateImpliedConstraints(relatedTo) {
    for (var _iterator = relatedTo.descendants, _isArray = Array.isArray(_iterator), _i = 0, _iterator = _isArray ? _iterator : _iterator[Symbol.iterator]();;) {
      var _ref;

      if (_isArray) {
        if (_i >= _iterator.length) break;
        _ref = _iterator[_i++];
      } else {
        _i = _iterator.next();
        if (_i.done) break;
        _ref = _i.value;
      }

      var baseObj = _ref;

      if (baseObj.isLine()) {
        // generate implied horizontality or verticality
        var vec = getTangent(baseObj.geomParams); // unit vector or zero

        if (vec.lengthSq() === 0) continue;
        if (Math.abs(vec.y) <= ConstraintGenerator.angularTolerance) this.adder.add(horizontality, [baseObj], {
          fixed: true,
          noVisualize: true
        });
        if (Math.abs(vec.x) <= ConstraintGenerator.angularTolerance) this.adder.add(verticality, [baseObj], {
          fixed: true,
          noVisualize: true
        });
      }

      if (!baseObj.isPoint()) continue;

      if (baseObj.geomParams.start.equals(new THREE.Vector3(0, 0, 0))) {
        this.adder.add(fixation, [baseObj]);
      }

      var basePar = baseObj.parent;

      for (var _iterator2 = this.sketch.descendants, _isArray2 = Array.isArray(_iterator2), _i2 = 0, _iterator2 = _isArray2 ? _iterator2 : _iterator2[Symbol.iterator]();;) {
        var _ref2;

        if (_isArray2) {
          if (_i2 >= _iterator2.length) break;
          _ref2 = _iterator2[_i2++];
        } else {
          _i2 = _iterator2.next();
          if (_i2.done) break;
          _ref2 = _i2.value;
        }

        var otherObj = _ref2;

        if (otherObj.isLine()) {
          var startOtherPos = otherObj.geomParams.start;
          var endOtherPos = otherObj.geomParams.end;
          var baseStart = baseObj.geomParams.start;
          if (startOtherPos.equals(baseStart) || endOtherPos.equals(baseStart)) //If curPoint match to line's point create constraint between points
            continue;
          var basePosVector = baseStart.clone().sub(startOtherPos);
          var lineVector = endOtherPos.clone().sub(startOtherPos);
          var crossVector = lineVector.clone().cross(basePosVector);
          var parallelogramArea = crossVector.length();
          var parallelogramHeight = lineVector.length(); // Distance from point to line is parallelogram area divided by parallelogram base

          var distanceToLine = parallelogramArea / parallelogramHeight; // If point on line, create coincidence constraint.

          if (distanceToLine === 0 && // If point on line (consider line width).
          Math.cos(basePosVector.angleTo(lineVector)) > 0 && // If point between startPoint and +infinity.
          basePosVector.length() <= parallelogramHeight // If point between -infinity and endPoint.
          ) this.adder.add(incidence, [baseObj, otherObj], {
              noVisualize: true
            });
          continue;
        }

        if (otherObj.isCircle() || otherObj.isArc()) {
          var _baseStart = baseObj.geomParams.start;
          var otherCenter = otherObj.geomParams.center;
          var otherRadius = otherObj.geomParams.radius;
          var diff = _baseStart.distanceTo(otherCenter) - otherRadius;

          if (Math.abs(diff) < ConstraintGenerator.linearTolerance) {
            if (otherObj.isCircle()) {
              this.adder.add(incidence, [baseObj, otherObj], {
                noVisualize: true
              });
            } else {
              var arcClockwise = otherObj.geomParams.clockwise; // If arc counterClockwise swap start and end to made arc clockwise.

              var otherStart = arcClockwise ? otherObj.geomParams.start : otherObj.geomParams.end;
              var otherEnd = arcClockwise ? otherObj.geomParams.end : otherObj.geomParams.start; // Calculate angles center-start, center-end, and center-base lines

              var startArcVector = otherStart.clone().sub(otherCenter);
              var endArcVector = otherEnd.clone().sub(otherCenter);

              var basePointVector = _baseStart.clone().sub(otherCenter);

              var startAngle = Math.atan2(startArcVector.y, startArcVector.x);
              var endAngle = Math.atan2(endArcVector.y, endArcVector.x);
              var baseAngle = Math.atan2(basePointVector.y, basePointVector.x); // Convert angles from (0, PI) or (0, -PI) intervals to (0, 2*PI) interval

              startAngle = startAngle < 0 ? 2 * Math.PI + startAngle : startAngle;
              endAngle = endAngle < 0 ? 2 * Math.PI + endAngle : endAngle;
              baseAngle = baseAngle < 0 ? 2 * Math.PI + baseAngle : baseAngle;
              var isOnArc = false;

              if (startAngle > endAngle) {
                isOnArc = baseAngle < startAngle && baseAngle > endAngle;
              } else {
                // Rotate all angles to made endAngle equals to zero
                var endToZeroDiff = 2 * Math.PI - endAngle;
                var newBase = baseAngle + endToZeroDiff;
                var newStart = startAngle + endToZeroDiff; // Working only in (0, 2*PI) interval.

                newBase = newBase > 2 * Math.PI ? newBase - 2 * Math.PI : newBase;
                newStart = newStart > 2 * Math.PI ? newStart - 2 * Math.PI : newStart;
                isOnArc = newBase < newStart && newBase > 0;
              }

              isOnArc && this.adder.add(incidence, [baseObj, otherObj], {
                noVisualize: true
              });
            }
          }

          continue;
        } // generate implied incidence of points


        if (!otherObj.isPoint()) continue;
        var basePos = baseObj.geomParams.start,
            otherPos = otherObj.geomParams.start;
        if (basePos.distanceTo(otherPos) > ConstraintGenerator.linearTolerance) continue;
        this.adder.add(incidence, [baseObj, otherObj], {
          noVisualize: true
        });
        var otherPar = otherObj.parent; // generate implied tangency of touching endpoints

        if (!baseObj.isSubPoint() || !otherObj.isSubPoint() || basePar.id === otherPar.id) continue;
        if (baseObj.isCenterPoint() || otherObj.isCenterPoint()) continue;
        var baseDir = getTangent(basePar.geomParams, basePos),
            otherDir = getTangent(otherPar.geomParams, otherPos);
        if (baseDir.lengthSq() === 0 || otherDir.lengthSq() === 0) continue;
        if (Math.abs(baseDir.cross(otherDir).z) > ConstraintGenerator.angularTolerance) continue;
        var type = basePar.isLine() && otherPar.isLine() ? incidence : tangency;
        this.adder.add(type, [basePar, otherPar], {
          noVisualize: true
        });
      }
    }

    return this.adder.commit();
  };

  return ConstraintGenerator;
}();

ConstraintGenerator.angularTolerance = 1e-6;
ConstraintGenerator.linearTolerance = 1e-6;
export { ConstraintGenerator as default };