import * as THREE from 'three';
import { LineSegmentsGeometry } from './LineSegmentsGeometry';
import { LineMaterial } from './LineMaterial';

function LineSegments2(geometry, material) {
  THREE.Mesh.call(this);
  this.type = 'LineSegments2';
  this.geometry = geometry !== undefined ? geometry : new LineSegmentsGeometry();
  this.material = material !== undefined ? material : new LineMaterial({
    color: Math.random() * 0xffffff
  });
}

LineSegments2.prototype = Object.assign(Object.create(THREE.Mesh.prototype), {
  constructor: LineSegments2,
  isLineSegments2: true,
  computeLineDistances: function () {
    // for backwards-compatability, but could be a method of LineSegmentsGeometry...
    var start = new THREE.Vector3();
    var end = new THREE.Vector3();
    return function computeLineDistances() {
      var geometry = this.geometry;
      var instanceStart = geometry.attributes.instanceStart;
      var instanceEnd = geometry.attributes.instanceEnd;
      var lineDistances = new Float32Array(2 * instanceStart.data.count);

      for (var i = 0, j = 0, l = instanceStart.data.count; i < l; i++, j += 2) {
        start.fromBufferAttribute(instanceStart, i);
        end.fromBufferAttribute(instanceEnd, i);
        lineDistances[j] = j === 0 ? 0 : lineDistances[j - 1];
        lineDistances[j + 1] = lineDistances[j] + start.distanceTo(end);
      }

      var instanceDistanceBuffer = new THREE.InstancedInterleavedBuffer(lineDistances, 2, 1); // d0, d1

      geometry.addAttribute('instanceDistanceStart', new THREE.InterleavedBufferAttribute(instanceDistanceBuffer, 1, 0)); // d0

      geometry.addAttribute('instanceDistanceEnd', new THREE.InterleavedBufferAttribute(instanceDistanceBuffer, 1, 1)); // d1

      return this;
    };
  }(),
  copy: function copy(source) {
    // todo
    return this;
  },
  raycast: function raycast(raycaster, intersects) {
    var view = raycaster.view;
    var geometry = this.geometry;
    var startPoint3 = new THREE.Vector3();
    var endPoint3 = new THREE.Vector3();
    startPoint3.fromBufferAttribute(geometry.attributes.instanceStart, 0);
    endPoint3.fromBufferAttribute(geometry.attributes.instanceEnd, 0);
    this.localToWorld(startPoint3);
    this.localToWorld(endPoint3); // Getting absolute dom pixel coordinates

    var startPoint2 = view.getPoint2(startPoint3.clone());
    var endPoint2 = view.getPoint2(endPoint3.clone()); // Sets depth to zero, startPoint2 and endPoint2 are 2d projections of startPoint3 and endPoint3 on DOM view

    startPoint2.setZ(0);
    endPoint2.setZ(0);
    var cursorVector = raycaster.interaction.clientCoord.clone().sub(startPoint2);
    var lineVector = endPoint2.clone().sub(startPoint2);
    var crossVector = lineVector.clone().cross(cursorVector);
    var parallelogramArea = crossVector.length();
    var parallelogramHeight = lineVector.length(); // Distance from point to line is parallelogram area divided by parallelogram base

    var distanceToLine = parallelogramArea / parallelogramHeight;

    if (distanceToLine <= this.material.linewidth && // If point on line (consider line width).
    Math.cos(cursorVector.angleTo(lineVector)) > 0 && // If point between startPoint and +infinity.
    cursorVector.length() <= parallelogramHeight // If point between -infinity and endPoint.
    ) {
        // Creating plane1 contains raycaster.ray.origin, startPoint3 and endPoint3.
        var startToOriginVec = raycaster.ray.origin.clone().sub(startPoint3);
        var startToEndVec = endPoint3.clone().sub(startPoint3); // Normal for plane1.

        var normal = new THREE.Vector3().crossVectors(startToOriginVec, startToEndVec); // Normal for plane2 (contains startPoint3, endPoint3 and perpendicular to plane1).

        normal.cross(startToEndVec); // Find intersection of ray.direction and plane2 .
        // Plane equation: Ax + By + Cz + d = 0.
        // Where (A, B, C) is (normal.x, normal.y, normal.z)
        // Ray parametric equation:
        // x = l * t + x1
        // y = m * t + y1
        // z = p * t + z1
        // Where (x1, y1, z1) is (ray.origin.x, ray.origin.y, ray.origin.z)
        // (l, m, p) is (ray.direction.x, ray.direction.y, ray.direction.z),
        // And t is parameter.

        var d = normal.dot(startPoint3.clone()); // Find parameter t with which ray intersects plane2.

        var t = (-1 * normal.dot(raycaster.ray.origin) + d) / normal.dot(raycaster.ray.direction); // Set t to ray parametric equation to find point.

        var point = raycaster.ray.direction.clone().multiplyScalar(t).add(raycaster.ray.origin);
        var intersection = {
          object: this,
          point: point,
          distance: raycaster.ray.origin.distanceTo(point)
        };
        intersects.push(intersection);
      }
  }
});
export { LineSegments2 };